﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2013 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.MainForm;

namespace HeuristicLab.DataPreprocessing.Views {
  [View("Data Grid Content View")]
  [Content(typeof(IDataGridContent), true)]
  public partial class DataGridContentView : CopyOfStringConvertibleMatrixView {

    private bool notOwnEvent = true;

    public new IDataGridContent Content {
      get { return (IDataGridContent)base.Content; }
      set { base.Content = value; }
    }

    private IList<int> _highlightedRowIndices;
    public IList<int> HighlightedRowIndices {
      get { return _highlightedRowIndices; }
      set {
        _highlightedRowIndices = value;
        Refresh();
      }
    }

    private IDictionary<int, IList<int>> _highlightedCells;
    public IDictionary<int, IList<int>> HightlightedCells {
      get { return _highlightedCells; }
      set {
        _highlightedCells = value;
        Refresh();
      }
    }

    public DataGridContentView() {
      InitializeComponent();
      dataGridView.CellMouseClick += dataGridView_CellMouseClick;
      dataGridView.CellPainting += new System.Windows.Forms.DataGridViewCellPaintingEventHandler(dataGridView_CellPainting);
      dataGridView.KeyUp += dataGridView_KeyUp;
      contextMenuCell.Items.Add(ShowHideColumns);
      _highlightedCells = new Dictionary<int, IList<int>>();
      _highlightedRowIndices = new List<int>();
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.Changed += Content_Changed;
    }

    protected override void DeregisterContentEvents() {
      base.DeregisterContentEvents();
      Content.Changed -= Content_Changed;
    }

    void Content_Changed(object sender, DataPreprocessingChangedEventArgs e) {
      if (notOwnEvent) {
        switch (e.Type) {
          case DataPreprocessingChangedEventType.AddColumn:
          case DataPreprocessingChangedEventType.AddRow:
          case DataPreprocessingChangedEventType.DeleteColumn:
          case DataPreprocessingChangedEventType.DeleteRow:
          case DataPreprocessingChangedEventType.Any:
            OnContentChanged();
            break;
          case DataPreprocessingChangedEventType.ChangeColumn:
          case DataPreprocessingChangedEventType.ChangeItem:
            dataGridView.Refresh();
            break;
        }

      }
    }

    protected override void dataGridView_CellValidating(object sender, DataGridViewCellValidatingEventArgs e) {
      if (!dataGridView.ReadOnly) {
        string errorMessage;
        if (Content != null && !Content.DataGridLogic.Validate(e.FormattedValue.ToString(), out errorMessage, e.ColumnIndex)) {
          e.Cancel = true;
          dataGridView.Rows[e.RowIndex].ErrorText = errorMessage;
        }
      }
    }

    protected override void dataGridView_CellParsing(object sender, DataGridViewCellParsingEventArgs e) {
      triggersOwnEvent(() => base.dataGridView_CellParsing(sender, e));
    }

    protected override void PasteValuesToDataGridView() {
      triggersOwnEvent(() => base.PasteValuesToDataGridView());
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      rowsTextBox.ReadOnly = true;
      columnsTextBox.ReadOnly = true;
    }

    private void btnApplySort_Click(object sender, System.EventArgs e) {
      triggersOwnEvent(() => {
        Content.PreprocessingDataManipulation.ReOrderToIndices(virtualRowIndices);
        OnContentChanged();
      });
    }

    private void triggersOwnEvent(Action action) {
      notOwnEvent = false;
      action();
      notOwnEvent = true;
    }

    private void dataGridView_KeyUp(object sender, KeyEventArgs e) {
      var selectedRows = dataGridView.SelectedRows;
      if (e.KeyCode == Keys.Delete && selectedRows.Count > 0) {
        List<int> rows = new List<int>();
        for (int i = 0; i < selectedRows.Count; ++i) {
          rows.Add(selectedRows[i].Index);
        }
        triggersOwnEvent(() => {
          Content.DataGridLogic.DeleteRow(rows);
          OnContentChanged();
        });
      }
    }

    private void dataGridView_CellMouseClick(object sender, DataGridViewCellMouseEventArgs e) {
      if (Content == null) return;
      if (e.Button == System.Windows.Forms.MouseButtons.Right) {
        if (e.ColumnIndex == -1 || e.RowIndex == -1) {
          contextMenu.Show(MousePosition);
        } else {
          if (!dataGridView.SelectedCells.Contains(dataGridView[e.ColumnIndex, e.RowIndex])) {
            dataGridView.ClearSelection();
            dataGridView[e.ColumnIndex, e.RowIndex].Selected = true;
          }
          interpolationToolStripMenuItem.Enabled = !(e.RowIndex == 0 || e.RowIndex == Content.Rows);
          var columnIndices = new HashSet<int>();
          for (int i = 0; i < dataGridView.SelectedCells.Count; i++) {
            columnIndices.Add(dataGridView.SelectedCells[i].ColumnIndex);
          }
          averageToolStripMenuItem.Enabled = medianToolStripMenuItem.Enabled = randomToolStripMenuItem.Enabled = !Content.DataGridLogic.AreAllStringColumns(columnIndices);
          interpolationToolStripMenuItem.Enabled = interpolationToolStripMenuItem.Enabled && !Content.DataGridLogic.AreAllStringColumns(columnIndices);
          contextMenuCell.Show(MousePosition);
        }
      }
    }

    protected void dataGridView_CellPainting(object sender, DataGridViewCellPaintingEventArgs e) {
      if (Content == null) return;
      if (e.RowIndex < 0) return;
      if (e.ColumnIndex < 0) return;
      if (e.State.HasFlag(DataGridViewElementStates.Selected)) return;
      if (!e.PaintParts.HasFlag(DataGridViewPaintParts.Background)) return;

      int rowIndex = virtualRowIndices[e.RowIndex];

      Color backColor = e.CellStyle.BackColor;

      if (HighlightedRowIndices != null && HighlightedRowIndices.Contains(rowIndex)
        || HightlightedCells != null && HightlightedCells.ContainsKey(e.ColumnIndex) && HightlightedCells[e.ColumnIndex].Contains(e.RowIndex)) {
        backColor = Color.Pink;
      }

      using (Brush backColorBrush = new SolidBrush(backColor)) {
        Rectangle bounds = new Rectangle(e.CellBounds.X, e.CellBounds.Y, e.CellBounds.Width, e.CellBounds.Height);
        e.Graphics.FillRectangle(backColorBrush, bounds);
      }

      using (Brush gridBrush = new SolidBrush(Color.LightGray)) {
        Pen gridLinePen = new Pen(gridBrush);
        e.Graphics.DrawLine(gridLinePen, e.CellBounds.Left,
               e.CellBounds.Bottom - 1, e.CellBounds.Right - 1,
               e.CellBounds.Bottom - 1);
        e.Graphics.DrawLine(gridLinePen, e.CellBounds.Right - 1,
            e.CellBounds.Top, e.CellBounds.Right - 1,
            e.CellBounds.Bottom);
      }

      e.PaintContent(e.CellBounds);
      e.Handled = true;
    }

    protected override int[] Sort(IEnumerable<KeyValuePair<int, SortOrder>> sortedColumns) {
      btnApplySort.Enabled = sortedColumns.Any();
      return base.Sort(sortedColumns);
    }

    protected override void ClearSorting() {
      btnApplySort.Enabled = false;
      base.ClearSorting();
    }

    private Dictionary<int, List<int>> GetSelectedCells() {
      var selectedCells = new Dictionary<int, List<int>>();
      for (int i = 0; i < dataGridView.SelectedCells.Count; i++) {
        var columnIndex = dataGridView.SelectedCells[i].ColumnIndex;
        if (!selectedCells.ContainsKey(columnIndex)) {
          selectedCells.Add(columnIndex, new List<int>());
        }
        selectedCells[columnIndex].Add(dataGridView.SelectedCells[i].RowIndex);
      }
      return selectedCells;
    }

    private void ReplaceWithAverage_Click(object sender, EventArgs e) {
      Content.PreprocessingDataManipulation.ReplaceIndicesByAverageValue(GetSelectedCells());
    }

    private void ReplaceWithMedian_Click(object sender, EventArgs e) {
      Content.PreprocessingDataManipulation.ReplaceIndicesByMedianValue(GetSelectedCells());
    }

    private void ReplaceWithRandom_Click(object sender, EventArgs e) {
      Content.PreprocessingDataManipulation.ReplaceIndicesByRandomValue(GetSelectedCells());
    }

    private void ReplaceWithMostCommon_Click(object sender, EventArgs e) {
      Content.PreprocessingDataManipulation.ReplaceIndicesByMostCommonValue(GetSelectedCells());
    }

    private void ReplaceWithInterpolation_Click(object sender, EventArgs e) {
      Content.PreprocessingDataManipulation.ReplaceIndicesByLinearInterpolationOfNeighbours(GetSelectedCells());
    }
  }
}
